#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
from neutron_lib import constants
from neutron_lib import context
from oslo_config import cfg
from oslo_utils import netutils
from oslo_utils import uuidutils

from neutron.ipam import driver
from neutron.ipam import exceptions as ipam_exc
from neutron.ipam import requests as ipam_req
from neutron import manager
from neutron.tests import base
from neutron.tests.unit.ipam import fake_driver

FAKE_IPAM_CLASS = 'neutron.tests.unit.ipam.fake_driver.FakeDriver'


class IpamSubnetRequestTestCase(base.BaseTestCase):

    def setUp(self):
        super(IpamSubnetRequestTestCase, self).setUp()
        self.tenant_id = uuidutils.generate_uuid()
        self.subnet_id = uuidutils.generate_uuid()


class TestIpamSubnetRequests(IpamSubnetRequestTestCase):

    def test_subnet_request(self):
        pool = ipam_req.SubnetRequest(self.tenant_id,
                                  self.subnet_id)
        self.assertEqual(self.tenant_id, pool.tenant_id)
        self.assertEqual(self.subnet_id, pool.subnet_id)
        self.assertIsNone(pool.gateway_ip)
        self.assertIsNone(pool.allocation_pools)

    def test_subnet_request_gateway(self):
        request = ipam_req.SubnetRequest(self.tenant_id,
                                     self.subnet_id,
                                     gateway_ip='1.2.3.1')
        self.assertEqual('1.2.3.1', str(request.gateway_ip))

    def test_subnet_request_bad_gateway(self):
        self.assertRaises(netaddr.core.AddrFormatError,
                          ipam_req.SubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          gateway_ip='1.2.3.')

    def test_subnet_request_with_range(self):
        allocation_pools = [netaddr.IPRange('1.2.3.4', '1.2.3.5'),
                            netaddr.IPRange('1.2.3.7', '1.2.3.9')]
        request = ipam_req.SubnetRequest(self.tenant_id,
                                     self.subnet_id,
                                     allocation_pools=allocation_pools)
        self.assertEqual(allocation_pools, request.allocation_pools)

    def test_subnet_request_range_not_list(self):
        self.assertRaises(TypeError,
                          ipam_req.SubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          allocation_pools=1)

    def test_subnet_request_bad_range(self):
        self.assertRaises(TypeError,
                          ipam_req.SubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          allocation_pools=['1.2.3.4'])

    def test_subnet_request_different_versions(self):
        pools = [netaddr.IPRange('0.0.0.1', '0.0.0.2'),
                 netaddr.IPRange('::1', '::2')]
        self.assertRaises(ValueError,
                          ipam_req.SubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          allocation_pools=pools)

    def test_subnet_request_overlap(self):
        pools = [netaddr.IPRange('0.0.0.10', '0.0.0.20'),
                 netaddr.IPRange('0.0.0.8', '0.0.0.10')]
        self.assertRaises(ValueError,
                          ipam_req.SubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          allocation_pools=pools)


class TestIpamAnySubnetRequest(IpamSubnetRequestTestCase):

    def test_subnet_request(self):
        request = ipam_req.AnySubnetRequest(self.tenant_id,
                                        self.subnet_id,
                                        constants.IPv4,
                                        24,
                                        gateway_ip='0.0.0.1')
        self.assertEqual(24, request.prefixlen)

    def test_subnet_request_bad_prefix_type(self):
        self.assertRaises(netaddr.core.AddrFormatError,
                          ipam_req.AnySubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          constants.IPv4,
                          'A')

    def test_subnet_request_bad_prefix(self):
        self.assertRaises(netaddr.core.AddrFormatError,
                          ipam_req.AnySubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          constants.IPv4,
                          33)
        self.assertRaises(netaddr.core.AddrFormatError,
                          ipam_req.AnySubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          constants.IPv6,
                          129)

    def test_subnet_request_gateway(self):
        request = ipam_req.AnySubnetRequest(self.tenant_id,
                                            self.subnet_id,
                                            constants.IPv6,
                                            64,
                                            gateway_ip='2000::1')
        self.assertEqual(netaddr.IPAddress('2000::1'), request.gateway_ip)

    def test_subnet_request_allocation_pool_wrong_version(self):
        pools = [netaddr.IPRange('0.0.0.4', '0.0.0.5')]
        self.assertRaises(ipam_exc.IpamValueInvalid,
                          ipam_req.AnySubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          constants.IPv6,
                          64,
                          allocation_pools=pools)

    def test_subnet_request_allocation_pool_not_in_net(self):
        pools = [netaddr.IPRange('0.0.0.64', '0.0.0.128')]
        self.assertRaises(ipam_exc.IpamValueInvalid,
                          ipam_req.AnySubnetRequest,
                          self.tenant_id,
                          self.subnet_id,
                          constants.IPv4,
                          25,
                          allocation_pools=pools)


class TestIpamSpecificSubnetRequest(IpamSubnetRequestTestCase):

    def test_subnet_request(self):
        request = ipam_req.SpecificSubnetRequest(self.tenant_id,
                                             self.subnet_id,
                                             '1.2.3.0/24',
                                             gateway_ip='1.2.3.1')
        self.assertEqual(24, request.prefixlen)
        self.assertEqual(netaddr.IPAddress('1.2.3.1'), request.gateway_ip)
        self.assertEqual(netaddr.IPNetwork('1.2.3.0/24'), request.subnet_cidr)

    def test_subnet_request_gateway(self):
        request = ipam_req.SpecificSubnetRequest(self.tenant_id,
                                                 self.subnet_id,
                                                 '2001::1',
                                                 gateway_ip='2000::1')
        self.assertEqual(netaddr.IPAddress('2000::1'), request.gateway_ip)


class TestAddressRequest(base.BaseTestCase):

    # This class doesn't test much.  At least running through all of the
    # constructors may shake out some trivial bugs.

    EUI64 = ipam_req.AutomaticAddressRequest.EUI64

    def test_specific_address_ipv6(self):
        request = ipam_req.SpecificAddressRequest('2000::45')
        self.assertEqual(netaddr.IPAddress('2000::45'), request.address)

    def test_specific_address_ipv4(self):
        request = ipam_req.SpecificAddressRequest('1.2.3.32')
        self.assertEqual(netaddr.IPAddress('1.2.3.32'), request.address)

    def test_any_address(self):
        ipam_req.AnyAddressRequest()

    def test_automatic_address_request_eui64(self):
        subnet_cidr = '2607:f0d0:1002:51::/64'
        port_mac = 'aa:bb:cc:dd:ee:ff'
        eui_addr = str(netutils.get_ipv6_addr_by_EUI64(subnet_cidr,
                                                       port_mac))
        request = ipam_req.AutomaticAddressRequest(
            address_type=self.EUI64,
            prefix=subnet_cidr,
            mac=port_mac)
        self.assertEqual(request.address, netaddr.IPAddress(eui_addr))

    def test_automatic_address_request_invalid_address_type_raises(self):
        self.assertRaises(ipam_exc.InvalidAddressType,
                          ipam_req.AutomaticAddressRequest,
                          address_type='kaboom')

    def test_automatic_address_request_eui64_no_mac_raises(self):
        self.assertRaises(ipam_exc.AddressCalculationFailure,
                          ipam_req.AutomaticAddressRequest,
                          address_type=self.EUI64,
                          prefix='meh')

    def test_automatic_address_request_eui64_alien_param_raises(self):
        self.assertRaises(ipam_exc.AddressCalculationFailure,
                          ipam_req.AutomaticAddressRequest,
                          address_type=self.EUI64,
                          mac='meh',
                          alien='et',
                          prefix='meh')


class TestIpamDriverLoader(base.BaseTestCase):

    def setUp(self):
        super(TestIpamDriverLoader, self).setUp()
        self.ctx = context.get_admin_context()

    def _verify_fake_ipam_driver_is_loaded(self, driver_name):
        mgr = manager.NeutronManager
        ipam_driver = mgr.load_class_for_provider('neutron.ipam_drivers',
                                                  driver_name)

        self.assertEqual(
            fake_driver.FakeDriver, ipam_driver,
            "loaded ipam driver should be FakeDriver")

    def _verify_import_error_is_generated(self, driver_name):
        mgr = manager.NeutronManager
        self.assertRaises(ImportError, mgr.load_class_for_provider,
                          'neutron.ipam_drivers',
                          driver_name)

    def test_ipam_driver_is_loaded_by_class(self):
        self._verify_fake_ipam_driver_is_loaded(FAKE_IPAM_CLASS)

    def test_ipam_driver_is_loaded_by_name(self):
        self._verify_fake_ipam_driver_is_loaded('fake')

    def test_ipam_driver_raises_import_error(self):
        self._verify_import_error_is_generated(
            'neutron.tests.unit.ipam_req.SomeNonExistentClass')

    def test_ipam_driver_raises_import_error_for_none(self):
        self._verify_import_error_is_generated(None)

    def _load_ipam_driver(self, driver_name, subnet_pool_id):
        cfg.CONF.set_override("ipam_driver", driver_name)
        return driver.Pool.get_instance(subnet_pool_id, self.ctx)

    def test_ipam_driver_is_loaded_from_ipam_driver_config_value(self):
        ipam_driver = self._load_ipam_driver('fake', None)
        self.assertIsInstance(
            ipam_driver, fake_driver.FakeDriver,
            "loaded ipam driver should be of type FakeDriver")

    @mock.patch(FAKE_IPAM_CLASS)
    def test_ipam_driver_is_loaded_with_subnet_pool_id(self, ipam_mock):
        subnet_pool_id = 'SomePoolID'
        self._load_ipam_driver('fake', subnet_pool_id)
        ipam_mock.assert_called_once_with(subnet_pool_id, self.ctx)


class TestAddressRequestFactory(base.BaseTestCase):

    def test_specific_address_request_is_loaded(self):
        for address in ('10.12.0.15', 'fffe::1'):
            ip = {'ip_address': address}
            port = {'device_owner':
                    constants.DEVICE_OWNER_COMPUTE_PREFIX + 'None'}
            self.assertIsInstance(
                ipam_req.AddressRequestFactory.get_request(None, port, ip),
                ipam_req.SpecificAddressRequest)

    def test_any_address_request_is_loaded(self):
        for addr in [None, '']:
            ip = {'ip_address': addr}
            port = {'device_owner':
                    constants.DEVICE_OWNER_COMPUTE_PREFIX + 'None'}
            self.assertIsInstance(
                ipam_req.AddressRequestFactory.get_request(None, port, ip),
                ipam_req.AnyAddressRequest)

    def test_automatic_address_request_is_loaded(self):
        ip = {'mac': '6c:62:6d:de:cf:49',
              'subnet_cidr': '2001:470:abcd::/64',
              'eui64_address': True}
        port = {'device_owner': constants.DEVICE_OWNER_COMPUTE_PREFIX + 'None'}
        self.assertIsInstance(
            ipam_req.AddressRequestFactory.get_request(None, port, ip),
            ipam_req.AutomaticAddressRequest)

    def test_prefernext_address_request_on_dhcp_port(self):
        ip = {}
        port = {'device_owner': constants.DEVICE_OWNER_DHCP}
        self.assertIsInstance(
            ipam_req.AddressRequestFactory.get_request(None, port, ip),
            ipam_req.PreferNextAddressRequest)

    def test_prefernext_address_request_on_distributed_port(self):
        ip = {}
        port = {'device_owner': constants.DEVICE_OWNER_DISTRIBUTED}
        self.assertIsInstance(
            ipam_req.AddressRequestFactory.get_request(None, port, ip),
            ipam_req.PreferNextAddressRequest)


class TestSubnetRequestFactory(IpamSubnetRequestTestCase):

    def _build_subnet_dict(self, id=None, cidr='192.168.1.0/24',
                           prefixlen=8, ip_version=constants.IP_VERSION_4,
                           gateway_ip=None):
        subnet = {'cidr': cidr,
                  'prefixlen': prefixlen,
                  'ip_version': ip_version,
                  'tenant_id': self.tenant_id,
                  'gateway_ip': gateway_ip,
                  'allocation_pools': None,
                  'id': id or self.subnet_id}
        subnetpool = {'ip_version': ip_version,
                      'default_prefixlen': prefixlen}
        return subnet, subnetpool

    def test_specific_subnet_request_is_loaded(self):
        addresses = [
            '10.12.0.15/24',
            '10.12.0.0/24',
            'fffe::1/64',
            'fffe::/64']
        for address in addresses:
            subnet, subnetpool = self._build_subnet_dict(cidr=address)
            self.assertIsInstance(
                ipam_req.SubnetRequestFactory.get_request(None,
                                                          subnet,
                                                          subnetpool),
                ipam_req.SpecificSubnetRequest)

    def test_specific_gateway_request_is_loaded(self):
        gw_prefixlen = [('10.12.0.15', 24), ('10.12.0.1', 8),
                        ('fffe::1', 64), ('fffe::', 64)]
        for gateway_ip, prefixlen in gw_prefixlen:
            subnet, subnetpool = self._build_subnet_dict(
                cidr=None, gateway_ip=gateway_ip, prefixlen=prefixlen)
            request = ipam_req.SubnetRequestFactory.get_request(
                None, subnet, subnetpool)

            cidr = netaddr.IPNetwork(str(gateway_ip) + '/%s' % prefixlen).cidr
            self.assertIsInstance(request, ipam_req.SpecificSubnetRequest)
            self.assertEqual(cidr, request.subnet_cidr)
            self.assertEqual(netaddr.IPAddress(gateway_ip), request.gateway_ip)
            self.assertEqual(prefixlen, request.prefixlen)

    def test_any_address_request_is_loaded_for_ipv4(self):
        subnet, subnetpool = self._build_subnet_dict(
            cidr=None, ip_version=constants.IP_VERSION_4)
        self.assertIsInstance(
            ipam_req.SubnetRequestFactory.get_request(None,
                                                      subnet,
                                                      subnetpool),
            ipam_req.AnySubnetRequest)

    def test_any_address_request_is_loaded_for_ipv6(self):
        subnet, subnetpool = self._build_subnet_dict(
            cidr=None, ip_version=constants.IP_VERSION_6)
        self.assertIsInstance(
            ipam_req.SubnetRequestFactory.get_request(None,
                                                      subnet,
                                                      subnetpool),
            ipam_req.AnySubnetRequest)

    def test_args_are_passed_to_specific_request(self):
        subnet, subnetpool = self._build_subnet_dict()
        request = ipam_req.SubnetRequestFactory.get_request(None,
                                                            subnet,
                                                            subnetpool)
        self.assertIsInstance(request,
                              ipam_req.SpecificSubnetRequest)
        self.assertEqual(self.tenant_id, request.tenant_id)
        self.assertEqual(self.subnet_id, request.subnet_id)
        self.assertIsNone(request.gateway_ip)
        self.assertIsNone(request.allocation_pools)


class TestGetRequestFactory(base.BaseTestCase):

    def setUp(self):
        super(TestGetRequestFactory, self).setUp()
        cfg.CONF.set_override('ipam_driver', 'fake')
        self.driver = driver.Pool.get_instance(None, None)

    def test_get_subnet_request_factory(self):
        self.assertEqual(
            self.driver.get_subnet_request_factory(),
            ipam_req.SubnetRequestFactory)

    def test_get_address_request_factory(self):
        self.assertEqual(
            self.driver.get_address_request_factory(),
            ipam_req.AddressRequestFactory)


class TestSubnetRequestMetaclass(base.BaseTestCase):

    def test__validate_gateway_ip_in_subnet(self):
        method = ipam_req.SubnetRequest._validate_gateway_ip_in_subnet
        cidr4 = netaddr.IPNetwork('192.168.0.0/24')
        self.assertIsNone(method(cidr4, cidr4.ip + 1))
        self.assertRaises(ipam_exc.IpamValueInvalid, method, cidr4, cidr4.ip)
        self.assertRaises(ipam_exc.IpamValueInvalid, method, cidr4,
                          cidr4.broadcast)

        cidr6 = netaddr.IPNetwork('2001:db8::/64')
        self.assertIsNone(method(cidr6, cidr6.ip + 1))
        self.assertIsNone(method(cidr6, cidr6.ip))
        self.assertIsNone(method(cidr6, cidr6.broadcast))



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base

DEPRECATED_REASON = """
The metering API now supports system scope and default roles.
"""

LABEL_COLLECTION_PATH = '/metering/metering-labels'
LABEL_RESOURCE_PATH = '/metering/metering-labels/{id}'

RULE_COLLECTION_PATH = '/metering/metering-label-rules'
RULE_RESOURCE_PATH = '/metering/metering-label-rules/{id}'


rules = [
    policy.DocumentedRuleDefault(
        name='create_metering_label',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Create a metering label',
        operations=[
            {
                'method': 'POST',
                'path': LABEL_COLLECTION_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_metering_label',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_metering_label',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        description='Get a metering label',
        operations=[
            {
                'method': 'GET',
                'path': LABEL_COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': LABEL_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_metering_label',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_metering_label',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Delete a metering label',
        operations=[
            {
                'method': 'DELETE',
                'path': LABEL_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_metering_label',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='create_metering_label_rule',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Create a metering label rule',
        operations=[
            {
                'method': 'POST',
                'path': RULE_COLLECTION_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='create_metering_label_rule',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='get_metering_label_rule',
        check_str=base.PROJECT_READER,
        scope_types=['project'],
        description='Get a metering label rule',
        operations=[
            {
                'method': 'GET',
                'path': RULE_COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': RULE_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='get_metering_label_rule',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
    policy.DocumentedRuleDefault(
        name='delete_metering_label_rule',
        check_str=base.PROJECT_ADMIN,
        scope_types=['project'],
        description='Delete a metering label rule',
        operations=[
            {
                'method': 'DELETE',
                'path': RULE_RESOURCE_PATH,
            },
        ],
        deprecated_rule=policy.DeprecatedRule(
            name='delete_metering_label_rule',
            check_str=base.RULE_ADMIN_ONLY,
            deprecated_reason=DEPRECATED_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    )
]


def list_rules():
    return rules



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib import context

from neutron.agent import resource_cache
from neutron.api.rpc.callbacks import events as events_rpc
from neutron.tests import base


class OVOLikeThing(object):
    def __init__(self, id, revision_number=10, **kwargs):
        self.id = id
        self.fields = ['id', 'revision_number']
        self.revision_number = revision_number
        for k, v in kwargs.items():
            self.fields.append(k)
            setattr(self, k, v)

    def to_dict(self):
        return {f: getattr(self, f) for f in self.fields}

    def get(self, k):
        return getattr(self, k, None)


class RemoteResourceCacheTestCase(base.BaseTestCase):
    def setUp(self):
        super(RemoteResourceCacheTestCase, self).setUp()
        rtypes = ['duck', 'goose']
        self.goose = OVOLikeThing(1)
        self.duck = OVOLikeThing(2)
        self.ctx = context.get_admin_context()
        self.rcache = resource_cache.RemoteResourceCache(rtypes)
        self._pullmock = mock.patch.object(self.rcache, '_puller').start()

    def test_get_resource_by_id(self):
        self.rcache.record_resource_update(self.ctx, 'goose', self.goose)
        self.assertEqual(self.goose,
                         self.rcache.get_resource_by_id('goose', 1))
        self.assertIsNone(self.rcache.get_resource_by_id('goose', 2))

    def test__flood_cache_for_query_pulls_once(self):
        resources = [OVOLikeThing(66), OVOLikeThing(67)]
        received_kw = []
        receiver = lambda r, e, t, payload: \
            received_kw.append(payload)
        registry.subscribe(receiver, 'goose', events.AFTER_UPDATE)

        self._pullmock.bulk_pull.side_effect = [
            resources,
            [resources[0]],
            [resources[1]],
            [resources[1]]
        ]

        self.rcache._flood_cache_for_query('goose', id=(66, 67),
                                           name=('a', 'b'))
        self._pullmock.bulk_pull.assert_called_once_with(
            mock.ANY, 'goose',
            filter_kwargs={'id': (66, 67), 'name': ('a', 'b')})

        self._pullmock.bulk_pull.reset_mock()
        self.rcache._flood_cache_for_query('goose', id=(66, ), name=('a', ))
        self.assertFalse(self._pullmock.called)
        self.rcache._flood_cache_for_query('goose', id=(67, ), name=('b', ))
        self.assertFalse(self._pullmock.called)

        # querying by just ID should trigger a new call since ID+name is a more
        # specific query
        self.rcache._flood_cache_for_query('goose', id=(67, ))
        self._pullmock.bulk_pull.assert_called_once_with(
            mock.ANY, 'goose', filter_kwargs={'id': (67, )})

        self.assertCountEqual(
            resources, [rec.latest_state for rec in received_kw])

    def test_bulk_pull_doesnt_wipe_out_newer_data(self):
        self.rcache.record_resource_update(
            self.ctx, 'goose', OVOLikeThing(1, revision_number=5))
        updated = OVOLikeThing(1)
        updated.revision_number = 1  # older revision number
        self._pullmock.bulk_pull.return_value = [updated]
        self.rcache._flood_cache_for_query('goose', id=(1,),)
        self.assertEqual(
             5, self.rcache.get_resource_by_id('goose', 1).revision_number)

    def test_get_resources(self):
        geese = [OVOLikeThing(3, size='large'), OVOLikeThing(5, size='medium'),
                 OVOLikeThing(4, size='large'), OVOLikeThing(6, size='small')]
        for goose in geese:
            self.rcache.record_resource_update(self.ctx, 'goose', goose)
        is_large = {'size': ('large', )}
        is_small = {'size': ('small', )}
        self.assertCountEqual([geese[0], geese[2]],
                              self.rcache.get_resources('goose', is_large))
        self.assertCountEqual([geese[3]],
                              self.rcache.get_resources('goose', is_small))

    def test_match_resources_with_func(self):
        geese = [OVOLikeThing(3, size='large'), OVOLikeThing(5, size='medium'),
                 OVOLikeThing(4, size='xlarge'), OVOLikeThing(6, size='small')]
        for goose in geese:
            self.rcache.record_resource_update(self.ctx, 'goose', goose)
        has_large = lambda o: 'large' in o.size
        self.assertCountEqual([geese[0], geese[2]],
                              self.rcache.match_resources_with_func('goose',
                                                                    has_large))

    def test__is_stale(self):
        goose = OVOLikeThing(3, size='large')
        self.rcache.record_resource_update(self.ctx, 'goose', goose)
        # same revision id is not considered stale
        updated = OVOLikeThing(3, size='large')
        self.assertFalse(self.rcache._is_stale('goose', updated))
        updated.revision_number = 0
        self.assertTrue(self.rcache._is_stale('goose', updated))
        updated.revision_number = 200
        self.assertFalse(self.rcache._is_stale('goose', updated))
        # once deleted, all updates are stale
        self.rcache.record_resource_delete(self.ctx, 'goose', 3)
        self.assertTrue(self.rcache._is_stale('goose', updated))

    def test_record_resource_update(self):
        received_kw = []
        receiver = lambda r, e, t, payload: \
            received_kw.append(payload)
        registry.subscribe(receiver, 'goose', events.AFTER_UPDATE)
        self.rcache.record_resource_update(self.ctx, 'goose',
                                           OVOLikeThing(3, size='large'))
        self.assertEqual(1, len(received_kw))
        self.assertIsNone(received_kw[0].states[0])
        # another update with no changed fields results in no callback
        self.rcache.record_resource_update(self.ctx, 'goose',
                                           OVOLikeThing(3, size='large',
                                                        revision_number=100))
        self.assertEqual(1, len(received_kw))
        self.rcache.record_resource_update(self.ctx, 'goose',
                                           OVOLikeThing(3, size='small',
                                                        revision_number=101))
        self.assertEqual(2, len(received_kw))
        self.assertEqual('large', received_kw[1].states[0].size)
        self.assertEqual('small', received_kw[1].latest_state.size)
        self.assertEqual(set(['size']),
                         received_kw[1].metadata['changed_fields'])

    def test_record_resource_delete(self):
        received_kw = []
        receiver = lambda r, e, t, payload: \
            received_kw.append(payload)
        registry.subscribe(receiver, 'goose', events.AFTER_DELETE)
        self.rcache.record_resource_update(self.ctx, 'goose',
                                           OVOLikeThing(3, size='large'))
        self.rcache.record_resource_delete(self.ctx, 'goose', 3)
        self.assertEqual(1, len(received_kw))
        self.assertEqual(3, received_kw[0].states[0].id)
        self.assertEqual(3, received_kw[0].resource_id)
        # deletes of non-existing cache items are still honored
        self.rcache.record_resource_delete(self.ctx, 'goose', 4)
        self.assertEqual(2, len(received_kw))
        self.assertIsNone(received_kw[1].states[0])
        self.assertEqual(4, received_kw[1].resource_id)

    def test_record_resource_delete_ignores_dups(self):
        received_kw = []
        receiver = lambda r, e, t, payload: \
            received_kw.append(payload)
        registry.subscribe(receiver, 'goose', events.AFTER_DELETE)
        self.rcache.record_resource_delete(self.ctx, 'goose', 3)
        self.assertEqual(1, len(received_kw))
        self.rcache.record_resource_delete(self.ctx, 'goose', 4)
        self.assertEqual(2, len(received_kw))
        self.rcache.record_resource_delete(self.ctx, 'goose', 3)
        self.assertEqual(2, len(received_kw))

    def test_resource_change_handler(self):
        with mock.patch.object(resource_cache.RemoteResourceWatcher,
                               '_init_rpc_listeners'):
            watch = resource_cache.RemoteResourceWatcher(self.rcache)
        geese = [OVOLikeThing(3, size='large'), OVOLikeThing(5, size='medium'),
                 OVOLikeThing(4, size='large'), OVOLikeThing(6, size='small')]
        watch.resource_change_handler(self.ctx, 'goose', geese,
                                      events_rpc.UPDATED)
        for goose in geese:
            self.assertEqual(goose,
                             self.rcache.get_resource_by_id('goose', goose.id))
        watch.resource_change_handler(self.ctx, 'goose', geese,
                                      events_rpc.DELETED)
        for goose in geese:
            self.assertIsNone(
                self.rcache.get_resource_by_id('goose', goose.id))



# Copyright (c) 2016 SUSE Linux Products GmbH
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import time

import eventlet
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import context as n_context
from neutron_lib.services.trunk import constants
from oslo_concurrency import lockutils
from oslo_context import context as o_context
from oslo_log import log as logging
import oslo_messaging
from oslo_serialization import jsonutils

from neutron._i18n import _
from neutron.agent.common import ovs_lib
from neutron.api.rpc.handlers import resources_rpc
from neutron.common import utils as common_utils
from neutron.plugins.ml2.drivers.openvswitch.agent.common \
    import constants as ovs_agent_constants
from neutron.services.trunk.drivers.openvswitch.agent import exceptions
from neutron.services.trunk.drivers.openvswitch.agent \
    import trunk_manager as tman
from neutron.services.trunk.drivers.openvswitch import constants as t_const
from neutron.services.trunk.drivers.openvswitch import utils
from neutron.services.trunk.rpc import agent

LOG = logging.getLogger(__name__)

DEFAULT_WAIT_FOR_PORT_TIMEOUT = 60
WAIT_BEFORE_TRUNK_DELETE = 6


def lock_on_bridge_name(required_parameter):
    def func_decor(f):
        try:
            br_arg_index = f.__code__.co_varnames.index(required_parameter)
        except ValueError:
            raise RuntimeError(_("%s parameter is required for this decorator")
                               % required_parameter)

        @functools.wraps(f)
        def inner(*args, **kwargs):
            try:
                bridge_name = kwargs[required_parameter]
            except KeyError:
                bridge_name = args[br_arg_index]
            with lockutils.lock(bridge_name):
                return f(*args, **kwargs)
        return inner
    return func_decor


def is_trunk_bridge(port_name):
    return port_name.startswith(t_const.TRUNK_BR_PREFIX)


def is_subport(port_name):
    return port_name.startswith(tman.SubPort.DEV_PREFIX)


def is_trunk_service_port(port_name):
    """True if the port is any of the ports used to realize a trunk."""
    return is_trunk_bridge(port_name) or port_name[:2] in (
        tman.TrunkParentPort.DEV_PREFIX,
        tman.SubPort.DEV_PREFIX)


def bridge_has_port(bridge, is_port_predicate):
    """True if there is an OVS port for which is_port_predicate is True.
    """
    try:
        ifaces = bridge.get_iface_name_list()
    except RuntimeError as e:
        LOG.error("Cannot obtain interface list for bridge %(bridge)s: "
                  "%(err)s",
                  {'bridge': bridge.br_name,
                   'err': e})
        return False

    return any(iface for iface in ifaces if is_port_predicate(iface))


def bridge_has_instance_port(bridge):
    """True if there is an OVS port that doesn't have bridge or patch ports
       prefix.
    """
    is_instance_port = lambda p: not is_trunk_service_port(p)
    return bridge_has_port(bridge, is_instance_port)


def bridge_has_service_port(bridge):
    """True if there is an OVS port that is used to implement a trunk.
    """
    return bridge_has_port(bridge, is_trunk_service_port)


@registry.has_registry_receivers
class OVSDBHandler(object):
    """It listens to OVSDB events to create the physical resources associated
    to a logical trunk in response to OVSDB events (such as VM boot and/or
    delete).
    """

    def __init__(self, trunk_manager):
        self.timeout = DEFAULT_WAIT_FOR_PORT_TIMEOUT
        self._context = n_context.get_admin_context_without_session()
        self.trunk_manager = trunk_manager
        self.trunk_rpc = agent.TrunkStub()

    @property
    def context(self):
        self._context.request_id = o_context.generate_request_id()
        return self._context

    @registry.receives(ovs_agent_constants.OVSDB_RESOURCE, [events.AFTER_READ])
    def process_trunk_port_events(
            self, resource, event, trigger, payload):
        """Process added and removed port events coming from OVSDB monitor."""
        ovsdb_events = payload.latest_state
        for port_event in ovsdb_events['added']:
            port_name = port_event['name']
            if is_trunk_bridge(port_name):
                LOG.debug("Processing trunk bridge %s", port_name)
                # As there is active waiting for port to appear, it's handled
                # in a separate greenthread.
                # NOTE: port_name is equal to bridge_name at this point.
                eventlet.spawn_n(self.handle_trunk_add, port_name)

        for port_event in ovsdb_events['removed']:
            bridge_name = port_event['external_ids'].get('bridge_name')
            if bridge_name and is_trunk_bridge(bridge_name):
                eventlet.spawn_n(
                    self.handle_trunk_remove, bridge_name, port_event)

    @lock_on_bridge_name(required_parameter='bridge_name')
    def handle_trunk_add(self, bridge_name):
        """Create trunk bridge based on parent port ID.

        This method is decorated with a lock that prevents processing deletion
        while creation hasn't been finished yet. It's based on the bridge name
        so we can keep processing other bridges in parallel.

        :param bridge_name: Name of the created trunk bridge.
        """
        bridge = ovs_lib.OVSBridge(bridge_name)
        # Handle condition when there was bridge in both added and removed
        # events and handle_trunk_remove greenthread was executed before
        # handle_trunk_add
        if not bridge.bridge_exists(bridge_name):
            LOG.debug("The bridge %s was deleted before it was handled.",
                      bridge_name)
            return

        # Determine the state of the trunk bridge by looking for the VM's port,
        # i.e. the trunk parent port and/or patch ports to be present. If the
        # VM is absent, then we clean the dangling bridge. If the VM is present
        # the value of 'rewire' tells us whether or not the bridge was dealt
        # with in a previous added event, and thus it has active patch ports.
        if not self._is_vm_connected(bridge):
            LOG.debug("No instance port associated to bridge %s could be "
                      "found. Deleting bridge and its resources.", bridge_name)
            self.trunk_manager.dispose_trunk(bridge)
            return

        # Check if the trunk was provisioned in a previous run. This can happen
        # at agent startup when existing trunks are notified as added events.
        rewire = bridge_has_service_port(bridge)
        # Once we get hold of the trunk parent port, we can provision
        # the OVS dataplane for the trunk.
        try:
            self._wire_trunk(bridge, self._get_parent_port(bridge), rewire)
        except oslo_messaging.MessagingException as e:
            LOG.error("Got messaging error while processing trunk bridge "
                      "%(bridge_name)s: %(err)s",
                      {'bridge_name': bridge.br_name,
                       'err': e})
        except exceptions.ParentPortNotFound as e:
            LOG.error("Failed to get parent port for bridge "
                      "%(bridge_name)s: %(err)s",
                      {'bridge_name': bridge.br_name,
                       'err': e})

    @lock_on_bridge_name(required_parameter='bridge_name')
    def handle_trunk_remove(self, bridge_name, port):
        """Remove wiring between trunk bridge and integration bridge.

        The method calls into trunk manager to remove patch ports on
        integration bridge side and to delete the trunk bridge. It's decorated
        with a lock to prevent deletion of bridge while creation is still in
        process.

        :param bridge_name: Name of the bridge used for locking purposes.
        :param port: Parent port dict.
        """
        # TODO(njohnston): In the case of DPDK with trunk ports, if nova
        # deletes an interface and then re-adds it we can get a race
        # condition where the port is re-added and then the bridge is
        # deleted because we did not properly catch the re-addition.  To
        # solve this would require transitioning to ordered event
        # resolution, like the L3 agent does with the
        # ResourceProcessingQueue class.  Until we can make that happen, we
        # try to mitigate the issue by checking if there is a port on the
        # bridge and if so then do not remove it.
        bridge = ovs_lib.OVSBridge(bridge_name)
        time.sleep(WAIT_BEFORE_TRUNK_DELETE)
        if bridge_has_instance_port(bridge):
            LOG.debug("The bridge %s has instances attached so it will not "
                      "be deleted.", bridge_name)
            return
        try:
            # TODO(jlibosva): Investigate how to proceed during removal of
            # trunk bridge that doesn't have metadata stored.
            parent_port_id, trunk_id, subport_ids = self._get_trunk_metadata(
                port)
            # NOTE(status_police): we do not report changes in trunk status on
            # removal to avoid potential races between agents in case the event
            # is due to a live migration or reassociation of a trunk to a new
            # VM.
            self.unwire_subports_for_trunk(trunk_id, subport_ids)
            self.trunk_manager.remove_trunk(trunk_id, parent_port_id)
        except tman.TrunkManagerError as te:
            LOG.error("Removing trunk %(trunk_id)s failed: %(err)s",
                      {'trunk_id': port['external_ids']['trunk_id'],
                       'err': te})
        else:
            LOG.debug("Deleted resources associated to trunk: %s", trunk_id)

    def manages_this_trunk(self, trunk_id):
        """True if this OVSDB handler manages trunk based on given ID."""
        bridge_name = utils.gen_trunk_br_name(trunk_id)
        return ovs_lib.BaseOVS().bridge_exists(bridge_name)

    def get_connected_subports_for_trunk(self, trunk_id):
        """Return the list of subports present on the trunk bridge."""
        bridge = ovs_lib.OVSBridge(utils.gen_trunk_br_name(trunk_id))
        if not bridge.bridge_exists(bridge.br_name):
            return []
        try:
            ports = bridge.get_ports_attributes(
                            'Interface', columns=['name', 'external_ids'])
            return [
                self.trunk_manager.get_port_uuid_from_external_ids(port)
                for port in ports if is_subport(port['name'])
            ]
        except (RuntimeError, tman.TrunkManagerError) as e:
            LOG.error("Failed to get subports for bridge %(bridge)s: "
                      "%(err)s", {'bridge': bridge.br_name, 'err': e})
            return []

    def wire_subports_for_trunk(self, context, trunk_id, subports,
                                trunk_bridge=None, parent_port=None):
        """Create OVS ports associated to the logical subports."""
        # Tell the server that subports must be bound to this host.
        subport_bindings = self.trunk_rpc.update_subport_bindings(
            context, subports)

        # Bindings were successful: create the OVS subports.
        subport_bindings = subport_bindings.get(trunk_id, [])
        subports_mac = {p['id']: p['mac_address'] for p in subport_bindings}
        subport_ids = []
        for subport in subports:
            try:
                self.trunk_manager.add_sub_port(trunk_id, subport.port_id,
                                                subports_mac[subport.port_id],
                                                subport.segmentation_id)
            except tman.TrunkManagerError as te:
                LOG.error("Failed to add subport with port ID "
                          "%(subport_port_id)s to trunk with ID "
                          "%(trunk_id)s: %(err)s",
                          {'subport_port_id': subport.port_id,
                           'trunk_id': trunk_id,
                           'err': te})
            else:
                subport_ids.append(subport.port_id)

        try:
            self._update_trunk_metadata(
                trunk_bridge, parent_port, trunk_id, subport_ids)
        except (RuntimeError, exceptions.ParentPortNotFound) as e:
            LOG.error("Failed to store metadata for trunk %(trunk_id)s: "
                      "%(reason)s", {'trunk_id': trunk_id, 'reason': e})
            # NOTE(status_police): Trunk bridge has stale metadata now, it
            # might cause troubles during deletion. Signal a DEGRADED status;
            # if the user undo/redo the operation things may go back to
            # normal.
            return constants.TRUNK_DEGRADED_STATUS

        LOG.debug("Added trunk: %s", trunk_id)
        return self._get_current_status(subports, subport_ids)

    def unwire_subports_for_trunk(self, trunk_id, subport_ids):
        """Destroy OVS ports associated to the logical subports."""
        ids = []
        for subport_id in subport_ids:
            try:
                self.trunk_manager.remove_sub_port(trunk_id, subport_id)
                ids.append(subport_id)
            except tman.TrunkManagerError as te:
                LOG.error("Removing subport %(subport_id)s from trunk "
                          "%(trunk_id)s failed: %(err)s",
                          {'subport_id': subport_id,
                           'trunk_id': trunk_id,
                           'err': te})
        try:
            # OVS bridge and port to be determined by _update_trunk_metadata
            bridge = None
            port = None
            self._update_trunk_metadata(
                bridge, port, trunk_id, subport_ids, wire=False)
        except RuntimeError as e:
            # NOTE(status_police): Trunk bridge has stale metadata now, it
            # might cause troubles during deletion. Signal a DEGRADED status;
            # if the user undo/redo the operation things may go back to
            # normal.
            LOG.error("Failed to store metadata for trunk %(trunk_id)s: "
                      "%(reason)s", {'trunk_id': trunk_id, 'reason': e})
            return constants.TRUNK_DEGRADED_STATUS
        except exceptions.ParentPortNotFound as e:
            # If a user deletes/migrates a VM and remove subports from a trunk
            # in short sequence, there is a chance that we hit this spot in
            # that the trunk may still be momentarily bound to the agent. We
            # should not mark the status as DEGRADED in this case.
            LOG.debug(e)

        return self._get_current_status(subport_ids, ids)

    def report_trunk_status(self, context, trunk_id, status):
        """Report trunk status to the server."""
        self.trunk_rpc.update_trunk_status(context, trunk_id, status)

    def _get_parent_port(self, trunk_bridge):
        """Return the OVS trunk parent port plugged on trunk_bridge."""
        trunk_br_ports = trunk_bridge.get_ports_attributes(
            'Interface', columns=['name', 'external_ids'],
            if_exists=True)
        for trunk_br_port in trunk_br_ports:
            if not is_trunk_service_port(trunk_br_port['name']):
                return trunk_br_port
        raise exceptions.ParentPortNotFound(bridge=trunk_bridge.br_name)

    def _wire_trunk(self, trunk_br, port, rewire=False):
        """Wire trunk bridge with integration bridge.

        The method calls into trunk manager to create patch ports for trunk and
        patch ports for all subports associated with this trunk. If rewire is
        True, a diff is performed between desired state (the one got from the
        server) and actual state (the patch ports present on the trunk bridge)
        and subports are wired/unwired accordingly.

        :param trunk_br: OVSBridge object representing the trunk bridge.
        :param port: Parent port dict.
        :param rewire: True if local trunk state must be reconciled with
            server's state.
        """
        ctx = self.context
        try:
            parent_port_id = (
                self.trunk_manager.get_port_uuid_from_external_ids(port))
            trunk = self.trunk_rpc.get_trunk_details(ctx, parent_port_id)
        except tman.TrunkManagerError:
            LOG.error("Can't obtain parent port ID from port %s",
                      port['name'])
            return
        except resources_rpc.ResourceNotFound:
            LOG.error("Port %s has no trunk associated.", parent_port_id)
            return

        try:
            registry.publish(
                resources.TRUNK, events.BEFORE_CREATE, self,
                payload=events.DBEventPayload(ctx, resource_id=trunk.id,
                                              desired_state=trunk))
            self.trunk_manager.create_trunk(
                trunk.id, trunk.port_id,
                port['external_ids'].get('attached-mac'))
        except tman.TrunkManagerError as te:
            LOG.error("Failed to create trunk %(trunk_id)s: %(err)s",
                      {'trunk_id': trunk.id,
                       'err': te})
            # NOTE(status_police): Trunk couldn't be created so it ends in
            # ERROR status and resync can fix that later.
            self.report_trunk_status(
                ctx, trunk.id, constants.TRUNK_ERROR_STATUS)
            return

        # We need to remove stale subports
        unwire_status = constants.TRUNK_ACTIVE_STATUS
        if rewire:
            old_subport_ids = self.get_connected_subports_for_trunk(trunk.id)
            subports = {p['port_id'] for p in trunk.sub_ports}
            subports_to_delete = set(old_subport_ids) - subports
            if subports_to_delete:
                unwire_status = self.unwire_subports_for_trunk(
                    trunk.id, subports_to_delete)

        # NOTE(status_police): inform the server whether the operation
        # was a partial or complete success. Do not inline status.
        # NOTE: in case of rewiring we readd ports that are already present on
        # the bridge because e.g. the segmentation ID might have changed (e.g.
        # agent crashed, port was removed and readded with a different seg ID)
        wire_status = self.wire_subports_for_trunk(
            ctx, trunk.id, trunk.sub_ports,
            trunk_bridge=trunk_br, parent_port=port)

        if (unwire_status == wire_status and
                wire_status == constants.TRUNK_ACTIVE_STATUS):
            status = constants.TRUNK_ACTIVE_STATUS
        else:
            status = constants.TRUNK_DEGRADED_STATUS
        self.report_trunk_status(ctx, trunk.id, status)

    def _set_trunk_metadata(self, trunk_bridge, port, trunk_id, subport_ids):
        """Set trunk metadata in OVS port for trunk parent port."""
        # update the parent port external_ids to store the trunk bridge
        # name, trunk id and subport ids so we can easily remove the trunk
        # bridge and service ports once this port is removed
        trunk_bridge = trunk_bridge or ovs_lib.OVSBridge(
            utils.gen_trunk_br_name(trunk_id))
        port = port or self._get_parent_port(trunk_bridge)

        port['external_ids']['bridge_name'] = trunk_bridge.br_name
        port['external_ids']['trunk_id'] = trunk_id
        port['external_ids']['subport_ids'] = jsonutils.dumps(subport_ids)
        trunk_bridge.set_db_attribute(
            'Interface', port['name'], 'external_ids', port['external_ids'])

    def _get_trunk_metadata(self, port):
        """Get trunk metadata from OVS port."""
        parent_port_id = (
            self.trunk_manager.get_port_uuid_from_external_ids(port))
        trunk_id = port['external_ids'].get('trunk_id')
        subport_ids = jsonutils.loads(
            port['external_ids'].get('subport_ids', '[]'))

        return parent_port_id, trunk_id, subport_ids

    def _update_trunk_metadata(self, trunk_bridge, port,
                               trunk_id, subport_ids, wire=True):
        """Update trunk metadata.

        :param trunk_bridge: OVS trunk bridge.
        :param port: OVS parent port.
        :param trunk_id: trunk ID.
        :param subport_ids: subports affecting the metadata.
        :param wire: if True subport_ids are added, otherwise removed.
        """
        trunk_bridge = trunk_bridge or ovs_lib.OVSBridge(
            utils.gen_trunk_br_name(trunk_id))
        port = port or self._get_parent_port(trunk_bridge)
        _port_id, _trunk_id, old_subports = self._get_trunk_metadata(port)
        if wire:
            new_subports = set(old_subports) | set(subport_ids)
        else:
            new_subports = set(old_subports) - set(subport_ids)
        self._set_trunk_metadata(trunk_bridge, port, trunk_id, new_subports)

    def _get_current_status(self, expected_subports, actual_subports):
        """Return the current status of the trunk.

        If the number of expected subports to be processed does not match the
        number of subports successfully processed, the status returned is
        DEGRADED, ACTIVE otherwise.
        """
        # NOTE(status_police): a call to this method should be followed by
        # a trunk_update_status to report the latest trunk status, but there
        # can be exceptions (e.g. unwire_subports_for_trunk).
        if len(expected_subports) != len(actual_subports):
            return constants.TRUNK_DEGRADED_STATUS
        else:
            return constants.TRUNK_ACTIVE_STATUS

    def _is_vm_connected(self, bridge):
        """True if an instance is connected to bridge, False otherwise."""
        bridge_has_port_predicate = functools.partial(
            bridge_has_instance_port, bridge)
        try:
            common_utils.wait_until_true(
                bridge_has_port_predicate,
                timeout=self.timeout)
            return True
        except common_utils.WaitTimeout:
            LOG.error(
                'No port present on trunk bridge %(br_name)s '
                'in %(timeout)d seconds.',
                {'br_name': bridge.br_name,
                 'timeout': self.timeout})
        return False



# Copyright (c) 2016 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.db import api as db_api
from oslo_utils import uuidutils

from neutron.db import l3_attrs_db
from neutron.objects.qos import binding as qos_binding
from neutron.objects.qos import policy
from neutron.objects import router
from neutron.tests.unit.objects import test_base as obj_test_base
from neutron.tests.unit import testlib_api


class RouterRouteIfaceObjectTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = router.RouterRoute


class RouterRouteDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                                  testlib_api.SqlTestCase):

    _test_class = router.RouterRoute

    def setUp(self):
        super(RouterRouteDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'router_id': lambda: self._create_test_router_id()})


class RouterExtraAttrsIfaceObjTestCase(obj_test_base.
                                       BaseObjectIfaceTestCase):
    _test_class = router.RouterExtraAttributes


class RouterExtraAttrsDbObjTestCase(obj_test_base.BaseDbObjectTestCase,
                                    testlib_api.SqlTestCase):
    _test_class = router.RouterExtraAttributes

    def setUp(self):
        super(RouterExtraAttrsDbObjTestCase, self).setUp()
        self.update_obj_fields(
            {'router_id': lambda: self._create_test_router_id()})


class RouterIfaceObjectTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = router.Router


class RouterDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                             testlib_api.SqlTestCase):

    _test_class = router.Router

    def setUp(self):
        super(RouterDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'gw_port_id': lambda: self._create_test_port_id(),
             'flavor_id': lambda: self._create_test_flavor_id()})

    def _create_router(self, router_id, gw_port_id, project_id):
        r = router.Router(self.context,
                          id=router_id,
                          gw_port_id=gw_port_id,
                          project_id=project_id)
        r.create()

    def test_check_routers_not_owned_by_projects(self):
        for obj in self.obj_fields:
            self._create_router(router_id=obj['id'],
                                gw_port_id=obj['gw_port_id'],
                                project_id=obj['project_id'])
        obj = self.obj_fields[0]

        gw_port = obj['gw_port_id']
        project = obj['project_id']
        new_project = project

        gw_port_no_match = uuidutils.generate_uuid()
        project_no_match = uuidutils.generate_uuid()
        new_project_no_match = uuidutils.generate_uuid()

        # Check router match with gw_port BUT no projects
        router_exist = router.Router.check_routers_not_owned_by_projects(
            self.context,
            [gw_port],
            [project_no_match, new_project_no_match])
        self.assertTrue(router_exist)

        # Check router doesn't match with gw_port
        router_exist = router.Router.check_routers_not_owned_by_projects(
            self.context,
            [gw_port_no_match],
            [project])
        self.assertFalse(router_exist)

        # Check router match with gw_port AND project
        router_exist = router.Router.check_routers_not_owned_by_projects(
            self.context,
            [gw_port],
            [project, new_project_no_match])
        self.assertFalse(router_exist)

        # Check router match with gw_port AND new project
        router_exist = router.Router.check_routers_not_owned_by_projects(
            self.context,
            [gw_port],
            [project_no_match, new_project])
        self.assertFalse(router_exist)

        # Check router match with gw_port AND project AND new project
        router_exist = router.Router.check_routers_not_owned_by_projects(
            self.context,
            [gw_port],
            [project, new_project])
        self.assertFalse(router_exist)

    def test_get_router_ids_without_router_std_attrs(self):
        def create_r_attr_reg(idx):
            with db_api.CONTEXT_WRITER.using(self.context):
                router_db = {'id': self.objs[idx].id}
                l3_attrs_db.ExtraAttributesMixin.add_extra_attr(self.context,
                                                                router_db)

        for idx in range(3):
            self.objs[idx].create()
        expected_router_ids = [r.id for r in self.objs]

        for idx in range(3):
            router_ids = router.Router.\
                get_router_ids_without_router_std_attrs(self.context)
            self.assertEqual(expected_router_ids, router_ids)
            create_r_attr_reg(idx)
            expected_router_ids = expected_router_ids[1:]


class RouterPortIfaceObjectTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = router.RouterPort


class RouterPortDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                                 testlib_api.SqlTestCase):

    _test_class = router.RouterPort

    def setUp(self):
        super(RouterPortDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'router_id': lambda: self._create_test_router_id(),
             'port_id': lambda: self._create_test_port_id()})


class DVRMacAddressIfaceObjectTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = router.DVRMacAddress


class DVRMacAddressDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                                    testlib_api.SqlTestCase):

    _test_class = router.DVRMacAddress


class FloatingIPIfaceObjectTestCase(obj_test_base.BaseObjectIfaceTestCase):

    _test_class = router.FloatingIP


class FloatingIPDbObjectTestCase(obj_test_base.BaseDbObjectTestCase,
                                 testlib_api.SqlTestCase):

    _test_class = router.FloatingIP

    def setUp(self):
        super(FloatingIPDbObjectTestCase, self).setUp()
        self.update_obj_fields(
            {'floating_port_id': lambda: self._create_test_port_id(),
             'fixed_port_id': lambda: self._create_test_port_id(),
             'router_id': lambda: self._create_test_router_id()})

    def test_qos_policy(self):
        _qos_policy_1 = self._create_test_qos_policy()
        _qos_policy_2 = self._create_test_qos_policy()

        self.obj_fields[0]['qos_policy_id'] = _qos_policy_1.id
        obj = self._test_class(
            self.context, **obj_test_base.remove_timestamps_from_fields(
                self.obj_fields[0], self._test_class.fields))
        obj.create()
        self.assertEqual(_qos_policy_1.id, obj.qos_policy_id)

        obj.qos_policy_id = _qos_policy_2.id
        obj.update()
        self.assertEqual(_qos_policy_2.id, obj.qos_policy_id)

        obj.qos_policy_id = None
        obj.update()
        self.assertIsNone(obj.qos_policy_id)

        obj.qos_policy_id = _qos_policy_1.id
        obj.update()
        fip_id = obj.id
        qos_fip_binding = qos_binding.QosPolicyFloatingIPBinding.get_objects(
            self.context, fip_id=fip_id)
        self.assertEqual(1, len(qos_fip_binding))
        self.assertEqual(_qos_policy_1.id, qos_fip_binding[0].policy_id)
        obj.delete()
        qos_fip_binding = qos_binding.QosPolicyFloatingIPBinding.get_objects(
            self.context, fip_id=fip_id)
        self.assertEqual([], qos_fip_binding)

    @mock.patch.object(policy.QosPolicy, 'unset_default')
    def test_qos_network_policy_id(self, *mocks):
        policy_obj = policy.QosPolicy(self.context)
        policy_obj.create()

        obj = self._make_object(self.obj_fields[0])
        obj.create()
        obj = router.FloatingIP.get_object(self.context, id=obj.id)
        self.assertIsNone(obj.qos_network_policy_id)
        self.assertIsNone(obj.qos_policy_id)

        network = self._create_test_network(qos_policy_id=policy_obj.id)
        self.update_obj_fields({'floating_network_id': network.id})
        obj = self._make_object(self.obj_fields[1])
        obj.create()
        obj = router.FloatingIP.get_object(self.context, id=obj.id)
        self.assertEqual(policy_obj.id, obj.qos_network_policy_id)
        self.assertIsNone(obj.qos_policy_id)

    def test_v1_1_to_v1_0_drops_qos_policy_id(self):
        obj = self._make_object(self.obj_fields[0])
        obj_v1_0 = obj.obj_to_primitive(target_version='1.0')
        self.assertNotIn('qos_policy_id', obj_v1_0['versioned_object.data'])

    def test_v1_2_to_v1_1_drops_qos_network_policy_id(self):
        obj = self._make_object(self.obj_fields[0])
        obj_v1_1 = obj.obj_to_primitive(target_version='1.1')
        self.assertNotIn('qos_network_policy_id',
                         obj_v1_1['versioned_object.data'])


class DvrFipGatewayPortAgentBindingTestCase(
        obj_test_base.BaseObjectIfaceTestCase):

    _test_class = router.DvrFipGatewayPortAgentBinding



# Copyright 2021 Huawei, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import netaddr
from neutron_lib.callbacks import events as lib_events
from neutron_lib.callbacks import registry as lib_registry
from neutron_lib import context
from os_ken.lib.packet import ether_types
from os_ken.lib.packet import in_proto as ip_proto
from oslo_utils import uuidutils

from neutron.agent.l2.extensions import local_ip as local_ip_ext
from neutron.api.rpc.callbacks import events
from neutron.api.rpc.callbacks import resources
from neutron.objects import local_ip as lip_obj
from neutron.plugins.ml2.drivers.openvswitch.agent.common import (
     constants as ovs_constants)
from neutron.plugins.ml2.drivers.openvswitch.agent \
    import ovs_agent_extension_api as ovs_ext_api
from neutron.tests import base


class LocalIPAgentExtensionTestCase(base.BaseTestCase):

    def setUp(self):
        super(LocalIPAgentExtensionTestCase, self).setUp()
        self.context = context.get_admin_context_without_session()
        self.local_ip_ext = local_ip_ext.LocalIPAgentExtension()

        self.plugin_rpc = mock.Mock()
        self.agent_api = ovs_ext_api.OVSAgentExtensionAPI(
            int_br=mock.Mock(),
            tun_br=mock.Mock(),
            phys_brs=None,
            plugin_rpc=self.plugin_rpc)
        self.local_ip_ext.consume_api(self.agent_api)
        with mock.patch.object(
                self.local_ip_ext, '_pull_all_local_ip_associations'):
            self.local_ip_ext.initialize(mock.Mock(), 'ovs')
        self.int_br = self.local_ip_ext.int_br

    def _generate_test_lip_associations(self, count=2):
        return [lip_obj.LocalIPAssociation(
            fixed_port_id=uuidutils.generate_uuid(),
            local_ip_id=uuidutils.generate_uuid(),
            fixed_ip='10.0.0.10',
            local_ip=lip_obj.LocalIP(
                local_ip_address='172.16.0.10')) for _ in range(count)]

    def test_pulling_lip_associations_on_init(self):
        res_rpc = mock.Mock()
        lip_assocs = self._generate_test_lip_associations()
        with mock.patch('neutron.api.rpc.handlers.'
                        'resources_rpc.ResourcesPullRpcApi') as res_rpc_cls:
            res_rpc_cls.return_value = res_rpc
            res_rpc.bulk_pull.return_value = lip_assocs
            self.local_ip_ext.initialize(mock.Mock(), 'ovs')

        res_rpc.bulk_pull.assert_called_once_with(
            mock.ANY, resources.LOCAL_IP_ASSOCIATION)

        for assoc in lip_assocs:
            self.assertEqual(
                assoc, self.local_ip_ext.local_ip_updates[
                    'added'][assoc.fixed_port_id][assoc.local_ip_id])

    def test_notify_port_updated(self):
        with mock.patch.object(lib_registry, "publish") as publish_mock:
            port_id = 'test'
            self.local_ip_ext._notify_port_updated(
                self.context, port_id=port_id)
            publish_mock.assert_called_once_with(
                resources.PORT, lib_events.AFTER_UPDATE,
                self.local_ip_ext, payload=mock.ANY)
            actual_payload = publish_mock.call_args[1]['payload']
            self.assertEqual(port_id, actual_payload.resource_id)
            self.assertEqual({'changed_fields': {'local_ip'}},
                             actual_payload.metadata)

    def test_handle_updated_notification(self):
        lip_assocs = self._generate_test_lip_associations()
        with mock.patch.object(
                self.local_ip_ext,
                "_notify_port_updated") as port_update_notify:
            self.local_ip_ext._handle_notification(
                self.context, resources.LOCAL_IP_ASSOCIATION,
                lip_assocs, events.UPDATED)

        for assoc in lip_assocs:
            self.assertEqual(
                assoc, self.local_ip_ext.local_ip_updates[
                    'added'][assoc.fixed_port_id][assoc.local_ip_id])
            port_update_notify.assert_any_call(
                self.context, assoc.fixed_port_id)

        return lip_assocs

    def test_handle_deleted_notification(self, lip_assocs=None):
        lip_assocs = lip_assocs or self.test_handle_updated_notification()
        with mock.patch.object(
                self.local_ip_ext,
                "_notify_port_updated") as port_update_notify:
            self.local_ip_ext._handle_notification(
                self.context, resources.LOCAL_IP_ASSOCIATION,
                lip_assocs, events.DELETED)
            for assoc in lip_assocs:
                self.assertEqual({}, self.local_ip_ext.local_ip_updates[
                    'added'][assoc.fixed_port_id])
                self.assertEqual(
                    assoc, self.local_ip_ext.local_ip_updates[
                        'deleted'][assoc.fixed_port_id][assoc.local_ip_id])
                port_update_notify.assert_any_call(
                    self.context, assoc.fixed_port_id)

    def test_handle_port(self):
        lip_assocs = self.test_handle_updated_notification()
        for assoc in lip_assocs:
            with mock.patch.object(self.local_ip_ext,
                                   'add_local_ip_flows') as add_lip_flows:
                port = {'port_id': assoc.fixed_port_id, 'local_vlan': 1}
                self.local_ip_ext.handle_port(self.context, port)
                self.assertEqual({}, self.local_ip_ext.local_ip_updates[
                    'added'][assoc.fixed_port_id])
                add_lip_flows.assert_called_once_with(port, assoc)
        self.test_handle_deleted_notification(lip_assocs)
        for assoc in lip_assocs:
            with mock.patch.object(self.local_ip_ext,
                                   'delete_local_ip_flows') as del_lip_flows:
                port = {'port_id': assoc.fixed_port_id, 'local_vlan': 1}
                self.local_ip_ext.handle_port(self.context, port)
                self.assertEqual({}, self.local_ip_ext.local_ip_updates[
                    'deleted'][assoc.fixed_port_id])
                del_lip_flows.assert_called_once_with(port, assoc)

    def test_delete_port(self):
        lip_assocs = self.test_handle_updated_notification()
        for assoc in lip_assocs:
            port = {'port_id': assoc.fixed_port_id}
            self.local_ip_ext.delete_port(self.context, port)

        self.assertEqual({}, self.local_ip_ext.local_ip_updates['added'])
        self.assertEqual({}, self.local_ip_ext.local_ip_updates['added'])

    def test_add_local_ip_flows(self):
        assoc = self._generate_test_lip_associations(1)[0]
        port = {'port_id': assoc.fixed_port_id,
                'mac_address': 'fa:16:3e:11:22:33',
                'local_vlan': 1234}
        with mock.patch.object(self.local_ip_ext,
                               'setup_local_ip_translation') as set_lip_trans:
            self.local_ip_ext.add_local_ip_flows(port, assoc)
            set_lip_trans.assert_called_once_with(
                vlan=port['local_vlan'],
                local_ip=str(assoc.local_ip.local_ip_address),
                dest_ip=str(assoc.fixed_ip),
                mac=port['mac_address']
            )
            self.int_br.install_arp_responder.assert_called_once_with(
                vlan=port['local_vlan'],
                ip=str(assoc.local_ip.local_ip_address),
                mac=port['mac_address'], table_id=31)
            self.int_br.install_garp_blocker.assert_called_once_with(
                vlan=port['local_vlan'],
                ip=str(assoc.local_ip.local_ip_address))
            self.int_br.install_garp_blocker_exception.assert_called_once_with(
                vlan=port['local_vlan'],
                ip=str(assoc.local_ip.local_ip_address),
                except_ip=str(assoc.fixed_ip))

    def test_delete_local_ip_flows(self):
        assoc = self._generate_test_lip_associations(1)[0]
        port = {'port_id': assoc.fixed_port_id,
                'mac_address': 'fa:16:3e:11:22:33',
                'local_vlan': 1234}
        with mock.patch.object(self.local_ip_ext,
                               'delete_local_ip_translation') as del_lip_trans:
            self.local_ip_ext.delete_local_ip_flows(port, assoc)
            del_lip_trans.assert_called_once_with(
                vlan=port['local_vlan'],
                local_ip=str(assoc.local_ip.local_ip_address),
                dest_ip=str(assoc.fixed_ip),
                mac=port['mac_address']
            )
            self.int_br.delete_arp_responder.assert_called_once_with(
                vlan=port['local_vlan'],
                ip=str(assoc.local_ip.local_ip_address),
                table_id=31)
            self.int_br.delete_garp_blocker.assert_called_once_with(
                vlan=port['local_vlan'],
                ip=str(assoc.local_ip.local_ip_address))
            self.int_br.delete_garp_blocker_exception.assert_called_once_with(
                vlan=port['local_vlan'],
                ip=str(assoc.local_ip.local_ip_address),
                except_ip=str(assoc.fixed_ip))

    def test_setup_local_ip_translation(self):
        vlan = 1234
        local_ip = '172.0.0.10'
        dest_ip = '10.0.0.10'
        mac = 'fa:16:3e:11:22:33'
        self.local_ip_ext.setup_local_ip_translation(
            vlan, local_ip, dest_ip, mac)

        expected_calls = [
            mock.call(
                table=31,
                priority=10,
                nw_dst=local_ip,
                reg6=vlan,
                dl_type="0x{:04x}".format(ether_types.ETH_TYPE_IP),
                actions='mod_dl_dst:{:s},'
                        'ct(commit,table={:d},zone={:d},nat(dst={:s}))'.format(
                    mac, 60, vlan, dest_ip)),
            mock.call(
                table=31,
                priority=10,
                dl_src=mac,
                nw_src=dest_ip,
                reg6=vlan,
                ct_state="-trk",
                dl_type="0x{:04x}".format(ether_types.ETH_TYPE_IP),
                actions='ct(table={:d},zone={:d},nat'.format(60, vlan)),
            mock.call(
                    table=31,
                    priority=11,
                    nw_src=dest_ip,
                    nw_dst=local_ip,
                    reg6=vlan,
                    dl_type="0x{:04x}".format(ether_types.ETH_TYPE_IP),
                    actions='resubmit(,{:d})'.format(60))
        ]
        self.assertEqual(expected_calls, self.int_br.add_flow.mock_calls)

    def test_delete_local_ip_translation(self):
        vlan = 1234
        local_ip = '172.0.0.10'
        dest_ip = '10.0.0.10'
        mac = 'fa:16:3e:11:22:33'
        self.local_ip_ext.delete_local_ip_translation(
            vlan, local_ip, dest_ip, mac)

        expected_calls = [
            mock.call(
                table_id=31,
                priority=10,
                ipv4_dst=local_ip,
                reg6=vlan,
                eth_type=ether_types.ETH_TYPE_IP),
            mock.call(
                table_id=31,
                priority=11,
                ipv4_src=dest_ip,
                ipv4_dst=local_ip,
                reg6=vlan,
                eth_type=ether_types.ETH_TYPE_IP),
            mock.call(
                table_id=31,
                priority=10,
                eth_src=mac,
                ipv4_src=dest_ip,
                reg6=vlan,
                eth_type=ether_types.ETH_TYPE_IP)
        ]
        self.assertEqual(
            expected_calls, self.int_br.uninstall_flows.mock_calls)

    def test_setup_static_local_ip_translation(self):
        ofpp_mock = mock.Mock()
        self.int_br._get_dp.return_value = (
            mock.Mock(), mock.Mock(), ofpp_mock)
        vlan = 1234
        local_ip = '172.0.0.10'
        dest_ip = '10.0.0.10'
        mac = 'fa:16:3e:11:22:33'
        self.local_ip_ext.setup_static_local_ip_translation(
            vlan, local_ip, dest_ip, mac)

        expected_calls = [
            mock.call(src=ether_types.ETH_TYPE_IP,
                      dst=('eth_type', 0), n_bits=16),
            mock.call(src=('eth_src', 0), dst=('eth_dst', 0), n_bits=48),
            mock.call(src=('eth_dst', 0), dst=('eth_src', 0), n_bits=48),
            mock.call(src=('ipv4_src', 0), dst=('ipv4_dst', 0), n_bits=32),
            mock.call(src=int(netaddr.IPAddress(dest_ip)),
                      dst=('ipv4_src', 0), n_bits=32),
            mock.call(src=vlan, dst=('reg6', 0), n_bits=4),
            mock.call(src=ip_proto.IPPROTO_ICMP, dst=('ip_proto', 0),
                      n_bits=8),
            mock.call(src=ip_proto.IPPROTO_TCP, dst=('ip_proto', 0),
                      n_bits=8),
            mock.call(src=('tcp_src', 0), dst=('tcp_dst', 0), n_bits=16),
            mock.call(src=('tcp_dst', 0), dst=('tcp_src', 0), n_bits=16),
            mock.call(src=ip_proto.IPPROTO_UDP, dst=('ip_proto', 0),
                      n_bits=8),
            mock.call(src=('udp_src', 0), dst=('udp_dst', 0), n_bits=16),
            mock.call(src=('udp_dst', 0), dst=('udp_src', 0), n_bits=16)
        ]
        self.assertEqual(
            expected_calls, ofpp_mock.NXFlowSpecMatch.mock_calls)

        ofpp_mock.NXFlowSpecLoad.assert_called_once_with(
            src=int(netaddr.IPAddress(local_ip)),
            dst=('ipv4_src', 0), n_bits=32)

        ofpp_mock.NXFlowSpecOutput.assert_called_once_with(
            src=('in_port', 0), dst='', n_bits=32)

        self.assertEqual(3, ofpp_mock.NXActionLearn.call_count)
        ofpp_mock.NXActionLearn.assert_called_with(
            table_id=ovs_constants.LOCAL_IP_TABLE,
            cookie=mock.ANY, priority=20, idle_timeout=30,
            hard_timeout=300, specs=mock.ANY)

        self.assertEqual(6, ofpp_mock.OFPActionSetField.call_count)
        ofpp_mock.OFPActionSetField.assert_any_call(ipv4_dst=dest_ip)
        ofpp_mock.OFPActionSetField.assert_any_call(eth_dst=mac)

        self.assertEqual(3, ofpp_mock.NXActionResubmitTable.call_count)
        ofpp_mock.NXActionResubmitTable.assert_called_with(
            table_id=ovs_constants.TRANSIENT_TABLE)

        self.assertEqual(3, self.int_br.install_apply_actions.call_count)
        self.int_br.install_apply_actions.assert_called_with(
            table_id=ovs_constants.LOCAL_IP_TABLE, match=mock.ANY,
            priority=10, actions=mock.ANY)

        self.int_br.add_flow.assert_called_once_with(
            table=31, priority=11, nw_src=dest_ip, nw_dst=local_ip,
            reg6=vlan, dl_type="0x{:04x}".format(ether_types.ETH_TYPE_IP),
            actions='resubmit(,{:d})'.format(60))



# Copyright (c) 2021 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from oslo_utils import uuidutils

from neutron.common import utils as common_utils
from neutron.tests.common.exclusive_resources import ip_network
from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment
from neutron.tests.fullstack.resources import machine


class OvsDHCPExtensionTestCase(base.BaseFullStackTestCase):
    number_of_hosts = 1

    def setUp(self):
        host_desc = [
            environment.HostDescription(
                l2_agent_type=constants.AGENT_TYPE_OVS,
                firewall_driver='openvswitch',
                # VM needs to receive the RA notification
                # from radvd which is handled by router and L3 agent.
                l3_agent=True,
                dhcp_agent=False) for _ in range(self.number_of_hosts)]
        env_desc = environment.EnvironmentDescription(
            mech_drivers='openvswitch',
            enable_traditional_dhcp=False)
        env = environment.Environment(env_desc, host_desc)
        super(OvsDHCPExtensionTestCase, self).setUp(env)
        self.tenant_id = uuidutils.generate_uuid()

        network = self.safe_client.create_network(
            self.tenant_id, name='public', external=True)
        cidr = self.useFixture(
            ip_network.ExclusiveIPNetwork(
                "240.0.0.0", "240.255.255.255", "24")).network
        self.safe_client.create_subnet(
            self.tenant_id, network['id'], cidr)

        router = self.safe_client.create_router(
            self.tenant_id, external_network=network['id'])

        self.network = self.safe_client.create_network(
            self.tenant_id, 'network-test')
        subnet_routes_v4 = [
            {"destination": "1.1.1.0/24", "nexthop": "10.0.0.100"},
            {"destination": "2.2.2.2/32", "nexthop": "10.0.0.101"}]
        self.subnet_v4 = self.safe_client.create_subnet(
            self.tenant_id, self.network['id'],
            cidr='10.0.0.0/24',
            gateway_ip='10.0.0.1',
            name='subnet-v4-test',
            host_routes=subnet_routes_v4)

        router_interface_info = self.safe_client.add_router_interface(
            router['id'], self.subnet_v4['id'])
        self.block_until_port_status_active(
            router_interface_info['port_id'])

        subnet_routes_v6 = [
            {"destination": "2001:4860:4860::8888/128",
             "nexthop": "fda7:a5cc:3460:1::1"},
            {"destination": "1234:5678:abcd::/64",
             "nexthop": "fda7:a5cc:3460:1::fff"}]
        self.subnet_v6 = self.safe_client.create_subnet(
            self.tenant_id, self.network['id'],
            cidr='fda7:a5cc:3460:1::/64',
            gateway_ip='fda7:a5cc:3460:1::1',
            enable_dhcp=True,
            ipv6_address_mode="dhcpv6-stateful",
            ipv6_ra_mode="dhcpv6-stateful",
            ip_version=6,
            name='subnet-v6-test',
            host_routes=subnet_routes_v6)

        # Need router radvd to send IPv6 address prefix to make the default
        # route work.
        router_interface_info = self.safe_client.add_router_interface(
            router['id'], self.subnet_v6['id'])
        self.block_until_port_status_active(
            router_interface_info['port_id'])

    def block_until_port_status_active(self, port_id):
        def is_port_status_active():
            port = self.client.show_port(port_id)
            return port['port']['status'] == 'ACTIVE'
        common_utils.wait_until_true(lambda: is_port_status_active(), sleep=1)

    def _prepare_vms(self):
        sgs = [self.safe_client.create_security_group(self.tenant_id)
               for _ in range(2)]

        port1 = self.safe_client.create_port(
            self.tenant_id, self.network['id'],
            self.environment.hosts[0].hostname,
            device_owner="compute:test_ovs_dhcp",
            security_groups=[sgs[0]['id']])

        port2 = self.safe_client.create_port(
            self.tenant_id, self.network['id'],
            self.environment.hosts[0].hostname,
            device_owner="compute:test_ovs_dhcp",
            security_groups=[sgs[1]['id']])

        # insert security-group-rules allow icmp
        self.safe_client.create_security_group_rule(
            self.tenant_id, sgs[0]['id'],
            direction=constants.INGRESS_DIRECTION,
            ethertype=constants.IPv4,
            protocol=constants.PROTO_NAME_ICMP)
        self.safe_client.create_security_group_rule(
            self.tenant_id, sgs[0]['id'],
            direction=constants.INGRESS_DIRECTION,
            ethertype=constants.IPv6,
            protocol=constants.PROTO_NAME_ICMP)

        # insert security-group-rules allow icmp
        self.safe_client.create_security_group_rule(
            self.tenant_id, sgs[1]['id'],
            direction=constants.INGRESS_DIRECTION,
            ethertype=constants.IPv4,
            protocol=constants.PROTO_NAME_ICMP)
        self.safe_client.create_security_group_rule(
            self.tenant_id, sgs[1]['id'],
            direction=constants.INGRESS_DIRECTION,
            ethertype=constants.IPv6,
            protocol=constants.PROTO_NAME_ICMP)

        vm1 = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.tenant_id,
                self.safe_client,
                neutron_port=port1,
                use_dhcp=True,
                use_dhcp6=True))

        vm2 = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.tenant_id,
                self.safe_client,
                neutron_port=port2,
                use_dhcp=True,
                use_dhcp6=True))
        return machine.FakeFullstackMachinesList([vm1, vm2])

    def test_ovs_dhcp_agent_extension_ping_vms(self):
        vms = self._prepare_vms()
        vms.block_until_all_boot()
        # ping -4 from vm_1 to vm_2
        vms.ping_all()
        # ping -6 from vm_1 to vm_2
        vms.ping6_all()



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.db import constants as db_const
from neutron_lib.db import model_base
import sqlalchemy as sa

from neutron.db import models_v2


class PortDeviceProfile(model_base.BASEV2):
    __tablename__ = 'portdeviceprofiles'
    port_id = sa.Column(sa.String(db_const.UUID_FIELD_SIZE),
                        sa.ForeignKey('ports.id', ondelete='CASCADE'),
                        primary_key=True)
    device_profile = sa.Column('device_profile', sa.String(length=255))
    port = sa.orm.relationship(
        models_v2.Port, load_on_pending=True,
        backref=sa.orm.backref('device_profile', uselist=False,
                               cascade='delete', lazy='joined'))

    revises_on_change = ('port', )



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import functools

from oslo_log import log as logging
from oslo_utils import excutils

from neutron.agent.linux import ip_lib

LOG = logging.getLogger(__name__)

NS_PREFIX = 'qrouter-'
INTERNAL_DEV_PREFIX = 'qr-'
EXTERNAL_DEV_PREFIX = 'qg-'
# TODO(Carl) It is odd that this file needs this.  It is a dvr detail.
ROUTER_2_FIP_DEV_PREFIX = 'rfp-'


def build_ns_name(prefix, identifier):
    """Builds a namespace name from the given prefix and identifier

    :param prefix: The prefix which must end with '-' for legacy reasons
    :param identifier: The id associated with the namespace
    """
    return prefix + identifier


def get_prefix_from_ns_name(ns_name):
    """Parses prefix from prefix-identifier

    :param ns_name: The name of a namespace
    :returns: The prefix ending with a '-' or None if there is no '-'
    """
    dash_index = ns_name.find('-')
    if 0 <= dash_index:
        return ns_name[:dash_index + 1]


def get_id_from_ns_name(ns_name):
    """Parses identifier from prefix-identifier

    :param ns_name: The name of a namespace
    :returns: Identifier or None if there is no - to end the prefix
    """
    dash_index = ns_name.find('-')
    if 0 <= dash_index:
        return ns_name[dash_index + 1:]


def check_ns_existence(f):
    @functools.wraps(f)
    def wrapped(self, *args, **kwargs):
        if not self.exists():
            LOG.warning('Namespace %(name)s does not exist. Skipping '
                        '%(func)s',
                        {'name': self.name, 'func': f.__name__})
            return
        try:
            return f(self, *args, **kwargs)
        except RuntimeError:
            with excutils.save_and_reraise_exception() as ctx:
                if not self.exists():
                    LOG.debug('Namespace %(name)s was concurrently deleted',
                              self.name)
                    ctx.reraise = False
    return wrapped


class Namespace(object):

    def __init__(self, name, agent_conf, driver, use_ipv6):
        self.name = name
        self.ip_wrapper_root = ip_lib.IPWrapper()
        self.agent_conf = agent_conf
        self.driver = driver
        self.use_ipv6 = use_ipv6

    def create(self, ipv6_forwarding=True):
        # See networking (netdev) tree, file
        # Documentation/networking/ip-sysctl.txt for an explanation of
        # these sysctl values.
        ip_wrapper = self.ip_wrapper_root.ensure_namespace(self.name)
        cmd = ['sysctl', '-w', 'net.ipv4.ip_forward=1']
        ip_wrapper.netns.execute(cmd, privsep_exec=True)
        # 1. Reply only if the target IP address is local address configured
        #    on the incoming interface; and
        # 2. Always use the best local address
        cmd = ['sysctl', '-w', 'net.ipv4.conf.all.arp_ignore=1']
        ip_wrapper.netns.execute(cmd, privsep_exec=True)
        cmd = ['sysctl', '-w', 'net.ipv4.conf.all.arp_announce=2']
        ip_wrapper.netns.execute(cmd, privsep_exec=True)
        if self.use_ipv6:
            cmd = ['sysctl', '-w',
                   'net.ipv6.conf.all.forwarding=%d' % int(ipv6_forwarding)]
            ip_wrapper.netns.execute(cmd, privsep_exec=True)

    def delete(self):
        try:
            self.ip_wrapper_root.netns.delete(self.name)
        except RuntimeError:
            msg = 'Failed trying to delete namespace: %s'
            LOG.exception(msg, self.name)

    def exists(self):
        return self.ip_wrapper_root.netns.exists(self.name)


class RouterNamespace(Namespace):

    def __init__(self, router_id, agent_conf, driver, use_ipv6):
        self.router_id = router_id
        name = self._get_ns_name(router_id)
        super(RouterNamespace, self).__init__(
            name, agent_conf, driver, use_ipv6)

    @classmethod
    def _get_ns_name(cls, router_id):
        return build_ns_name(NS_PREFIX, router_id)

    @check_ns_existence
    def delete(self):
        ns_ip = ip_lib.IPWrapper(namespace=self.name)
        for d in ns_ip.get_devices():
            if d.name.startswith(INTERNAL_DEV_PREFIX):
                # device is on default bridge
                self.driver.unplug(d.name, namespace=self.name,
                                   prefix=INTERNAL_DEV_PREFIX)
            elif d.name.startswith(ROUTER_2_FIP_DEV_PREFIX):
                ns_ip.del_veth(d.name)
            elif d.name.startswith(EXTERNAL_DEV_PREFIX):
                self.driver.unplug(
                    d.name,
                    namespace=self.name,
                    prefix=EXTERNAL_DEV_PREFIX)

        super(RouterNamespace, self).delete()



# Copyright 2016 GoDaddy.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#  implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import netaddr
from neutron_lib import constants

import neutron.api.extensions as api_ext
import neutron.common.config as config
import neutron.extensions
import neutron.services.network_ip_availability.plugin as plugin_module
import neutron.tests.unit.db.test_db_base_plugin_v2 as test_db_base_plugin_v2

API_RESOURCE = 'network-ip-availabilities'
IP_AVAIL_KEY = 'network_ip_availability'
IP_AVAILS_KEY = 'network_ip_availabilities'
EXTENSIONS_PATH = ':'.join(neutron.extensions.__path__)
PLUGIN_NAME = '%s.%s' % (plugin_module.NetworkIPAvailabilityPlugin.__module__,
                         plugin_module.NetworkIPAvailabilityPlugin.__name__)


class TestNetworkIPAvailabilityAPI(
        test_db_base_plugin_v2.NeutronDbPluginV2TestCase):
    def setUp(self):
        svc_plugins = {'plugin_name': PLUGIN_NAME}
        super(TestNetworkIPAvailabilityAPI, self).setUp(
                service_plugins=svc_plugins)
        self.plugin = plugin_module.NetworkIPAvailabilityPlugin()
        ext_mgr = api_ext.PluginAwareExtensionManager(
            EXTENSIONS_PATH, {"network-ip-availability": self.plugin}
        )
        app = config.load_paste_app('extensions_test_app')
        self.ext_api = api_ext.ExtensionMiddleware(app, ext_mgr=ext_mgr)

    def _validate_availability(self, network, availability, expected_used_ips,
                               expected_total_ips=253):
        self.assertEqual(network['name'], availability['network_name'])
        self.assertEqual(network['id'], availability['network_id'])
        self.assertEqual(expected_used_ips, availability['used_ips'])
        self.assertEqual(expected_total_ips, availability['total_ips'])

    def _validate_from_availabilities(self, availabilities, wrapped_network,
                                      expected_used_ips,
                                      expected_total_ips=253):
        network = wrapped_network['network']
        availability = self._find_availability(availabilities, network['id'])
        self.assertIsNotNone(availability)
        self._validate_availability(network, availability,
                                    expected_used_ips=expected_used_ips,
                                    expected_total_ips=expected_total_ips)

    def test_usages_query_list_with_fields_total_ips(self):
        with self.network() as net:
            with self.subnet(network=net):
                # list by query fields: total_ips
                params = 'fields=total_ips'
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertIn(IP_AVAILS_KEY, response)
                self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                availability = response[IP_AVAILS_KEY][0]
                self.assertIn('total_ips', availability)
                self.assertEqual(253, availability['total_ips'])
                self.assertNotIn('network_id', availability)

    def test_usages_query_show_with_fields_total_ips(self):
        with self.network() as net:
            with self.subnet(network=net):
                network = net['network']
                # Show by query fields: total_ips
                params = ['total_ips']
                request = self.new_show_request(API_RESOURCE,
                                                network['id'],
                                                fields=params)
                response = self.deserialize(
                    self.fmt, request.get_response(self.ext_api))
                self.assertIn(IP_AVAIL_KEY, response)
                availability = response[IP_AVAIL_KEY]
                self.assertIn('total_ips', availability)
                self.assertEqual(253, availability['total_ips'])
                self.assertNotIn('network_id', availability)

    @staticmethod
    def _find_availability(availabilities, net_id):
        for ip_availability in availabilities:
            if net_id == ip_availability['network_id']:
                return ip_availability

    def test_basic(self):
        with self.network() as net:
            with self.subnet(network=net):
                network = net['network']
                # Get ALL
                request = self.new_list_request(API_RESOURCE, self.fmt)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertIn(IP_AVAILS_KEY, response)
                self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                   net, 0)

                # Get single via id
                request = self.new_show_request(API_RESOURCE, network['id'])
                response = self.deserialize(
                    self.fmt, request.get_response(self.ext_api))
                self.assertIn(IP_AVAIL_KEY, response)
                usage = response[IP_AVAIL_KEY]
                self._validate_availability(network, usage, 0)

    def test_usages_multi_nets_subnets(self):
        with self.network(name='net1') as n1,\
                self.network(name='net2') as n2,\
                self.network(name='net3') as n3:
            # n1 should have 2 subnets, n2 should have none, n3 has 1
            with self.subnet(network=n1) as subnet1_1, \
                    self.subnet(cidr='40.0.0.0/24', network=n3) as subnet3_1:
                # Consume 3 ports n1, none n2, 2 ports on n3
                with self.port(subnet=subnet1_1),\
                        self.port(subnet=subnet1_1),\
                        self.port(subnet=subnet1_1),\
                        self.port(subnet=subnet3_1),\
                        self.port(subnet=subnet3_1):

                    # Test get ALL
                    request = self.new_list_request(API_RESOURCE)
                    response = self.deserialize(
                        self.fmt, request.get_response(self.ext_api))
                    self.assertIn(IP_AVAILS_KEY, response)
                    self.assertEqual(3, len(response[IP_AVAILS_KEY]))

                    data = response[IP_AVAILS_KEY]
                    self._validate_from_availabilities(data, n1, 3, 253)
                    self._validate_from_availabilities(data, n2, 0, 0)
                    self._validate_from_availabilities(data, n3, 2, 253)

                    # Test get single via network id
                    network = n1['network']
                    request = self.new_show_request(API_RESOURCE,
                                                    network['id'])
                    response = self.deserialize(
                        self.fmt, request.get_response(self.ext_api))
                    self.assertIn(IP_AVAIL_KEY, response)
                    self._validate_availability(network,
                                                response[IP_AVAIL_KEY], 3, 253)

    def test_usages_multi_nets_subnets_sums(self):
        with self.network(name='net1') as n1:
            # n1 has 2 subnets
            with self.subnet(network=n1) as subnet1_1, \
                    self.subnet(cidr='40.0.0.0/24', network=n1) as subnet1_2:
                # Consume 3 ports n1: 1 on subnet 1 and 2 on subnet 2
                with self.port(subnet=subnet1_1),\
                        self.port(subnet=subnet1_2),\
                        self.port(subnet=subnet1_2):
                    # Get ALL
                    request = self.new_list_request(API_RESOURCE)
                    response = self.deserialize(
                        self.fmt, request.get_response(self.ext_api))
                    self.assertIn(IP_AVAILS_KEY, response)
                    self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                    self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                       n1, 3, 506)

                    # Get single via network id
                    network = n1['network']
                    request = self.new_show_request(API_RESOURCE,
                                                    network['id'])
                    response = self.deserialize(
                        self.fmt, request.get_response(self.ext_api))
                    self.assertIn(IP_AVAIL_KEY, response)
                    self._validate_availability(network,
                                                response[IP_AVAIL_KEY], 3, 506)

    def test_usages_port_consumed_v4(self):
        with self.network() as net:
            with self.subnet(network=net) as subnet:
                request = self.new_list_request(API_RESOURCE)
                # Consume 2 ports
                with self.port(subnet=subnet), self.port(subnet=subnet):
                    response = self.deserialize(self.fmt,
                                                request.get_response(
                                                    self.ext_api))
                    self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                       net, 2)

    def test_usages_query_ip_version_v4(self):
        with self.network() as net:
            with self.subnet(network=net):
                # Get IPv4
                params = 'ip_version=%s' % constants.IP_VERSION_4
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertIn(IP_AVAILS_KEY, response)
                self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                   net, 0)

                # Get IPv6 should return empty array
                params = 'ip_version=%s' % constants.IP_VERSION_6
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertEqual(0, len(response[IP_AVAILS_KEY]))

    def test_usages_query_ip_version_v6(self):
        cidr_ipv6 = '2001:db8:1002:51::/64'
        cidr_ipv6_net = netaddr.IPNetwork(cidr_ipv6)
        with self.network() as net:
            with self.subnet(
                    network=net, cidr=cidr_ipv6,
                    ip_version=constants.IP_VERSION_6,
                    ipv6_address_mode=constants.DHCPV6_STATELESS):
                # Get IPv6
                params = 'ip_version=%s' % constants.IP_VERSION_6
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                self._validate_from_availabilities(
                    response[IP_AVAILS_KEY], net, 0, cidr_ipv6_net.size - 1)

                # Get IPv4 should return empty array
                params = 'ip_version=%s' % constants.IP_VERSION_4
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertEqual(0, len(response[IP_AVAILS_KEY]))

    def test_usages_ports_consumed_v6(self):
        cidr_ipv6 = '2001:db8:1002:51::/64'
        cidr_ipv6_net = netaddr.IPNetwork(cidr_ipv6)
        with self.network() as net:
            with self.subnet(
                    network=net, cidr=cidr_ipv6,
                    ip_version=constants.IP_VERSION_6,
                    ipv6_address_mode=constants.DHCPV6_STATELESS) as subnet:
                request = self.new_list_request(API_RESOURCE)
                # Consume 3 ports
                with self.port(subnet=subnet),\
                        self.port(subnet=subnet), \
                        self.port(subnet=subnet):
                    response = self.deserialize(
                        self.fmt, request.get_response(self.ext_api))

                    self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                       net, 3,
                                                       cidr_ipv6_net.size - 1)

    def test_usages_query_network_id(self):
        with self.network() as net:
            with self.subnet(network=net):
                network = net['network']
                test_id = network['id']
                # Get by query param: network_id
                params = 'network_id=%s' % test_id
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertIn(IP_AVAILS_KEY, response)
                self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                   net, 0)

                # Get by NON-matching query param: network_id
                params = 'network_id=clearlywontmatch'
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertEqual(0, len(response[IP_AVAILS_KEY]))

    def test_usages_query_network_name(self):
        test_name = 'net_name_1'
        with self.network(name=test_name) as net:
            with self.subnet(network=net):
                # Get by query param: network_name
                params = 'network_name=%s' % test_name
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertIn(IP_AVAILS_KEY, response)
                self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                   net, 0)

                # Get by NON-matching query param: network_name
                params = 'network_name=clearly-wont-match'
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertEqual(0, len(response[IP_AVAILS_KEY]))

    def test_usages_query_tenant_id(self):
        test_tenant_id = 'a-unique-test-id'
        with self.network(tenant_id=test_tenant_id) as net:
            with self.subnet(network=net):
                # Get by query param: tenant_id
                params = 'tenant_id=%s' % test_tenant_id
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertIn(IP_AVAILS_KEY, response)
                self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                   net, 0)
                for net_avail in response[IP_AVAILS_KEY]:
                    self.assertEqual(test_tenant_id, net_avail['tenant_id'])

                # Get by NON-matching query param: tenant_id
                params = 'tenant_id=clearly-wont-match'
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertEqual(0, len(response[IP_AVAILS_KEY]))

    def test_usages_query_project_id(self):
        test_project_id = 'a-unique-project-id'
        with self.network(tenant_id=test_project_id) as net:
            with self.subnet(network=net):
                # Get by query param: project_id
                params = 'project_id=%s' % test_project_id
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertIn(IP_AVAILS_KEY, response)
                self.assertEqual(1, len(response[IP_AVAILS_KEY]))
                self._validate_from_availabilities(response[IP_AVAILS_KEY],
                                                   net, 0)
                for net_avail in response[IP_AVAILS_KEY]:
                    self.assertEqual(test_project_id, net_avail['project_id'])

                # Get by NON-matching query param: project_id
                params = 'project_id=clearly-wont-match'
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertEqual(0, len(response[IP_AVAILS_KEY]))

    def test_usages_multi_net_multi_subnet_46(self):
        # Setup mixed v4/v6 networks with IPs consumed on each
        cidr_ipv6 = '2001:db8:1003:52::/64'
        cidr_ipv6_net = netaddr.IPNetwork(cidr_ipv6)
        with self.network(name='net-v6-1') as net_v6_1, \
                self.network(name='net-v6-2') as net_v6_2, \
                self.network(name='net-v4-1') as net_v4_1, \
                self.network(name='net-v4-2') as net_v4_2:
            with self.subnet(network=net_v6_1, cidr='2001:db8:1002:51::/64',
                             ip_version=constants.IP_VERSION_6) as s61, \
                    self.subnet(network=net_v6_2,
                                cidr=cidr_ipv6,
                                ip_version=constants.IP_VERSION_6) as s62, \
                    self.subnet(network=net_v4_1, cidr='10.0.0.0/24') as s41, \
                    self.subnet(network=net_v4_2, cidr='10.0.1.0/24') as s42:
                with self.port(subnet=s61),\
                        self.port(subnet=s62), self.port(subnet=s62), \
                        self.port(subnet=s41), \
                        self.port(subnet=s42), self.port(subnet=s42):

                    # Verify consumption across all
                    request = self.new_list_request(API_RESOURCE)
                    response = self.deserialize(
                        self.fmt, request.get_response(self.ext_api))
                    avails_list = response[IP_AVAILS_KEY]
                    self._validate_from_availabilities(
                            avails_list, net_v6_1, 1, cidr_ipv6_net.size - 1)
                    self._validate_from_availabilities(
                            avails_list, net_v6_2, 2, cidr_ipv6_net.size - 1)
                    self._validate_from_availabilities(
                            avails_list, net_v4_1, 1, 253)
                    self._validate_from_availabilities(
                            avails_list, net_v4_2, 2, 253)

                    # Query by IP versions. Ensure subnet versions match
                    for ip_ver in [constants.IP_VERSION_4,
                                   constants.IP_VERSION_6]:
                        params = 'ip_version=%i' % ip_ver
                        request = self.new_list_request(API_RESOURCE,
                                                        params=params)
                        response = self.deserialize(
                                self.fmt, request.get_response(self.ext_api))
                        for net_avail in response[IP_AVAILS_KEY]:
                            for sub in net_avail['subnet_ip_availability']:
                                self.assertEqual(ip_ver, sub['ip_version'])

                    # Verify consumption querying 2 network ids (IN clause)
                    request = self.new_list_request(
                            API_RESOURCE,
                            params='network_id=%s&network_id=%s'
                                   % (net_v4_2['network']['id'],
                                      net_v6_2['network']['id']))
                    response = self.deserialize(
                        self.fmt, request.get_response(self.ext_api))
                    avails_list = response[IP_AVAILS_KEY]
                    self._validate_from_availabilities(
                            avails_list, net_v6_2, 2, cidr_ipv6_net.size - 1)
                    self._validate_from_availabilities(
                            avails_list, net_v4_2, 2, 253)

    def test_usages_query_limit(self):
        with self.network() as net1, self.network() as net2, \
                self.network() as net3, self.network() as net4:
            networks = (net1, net2, net3, net4)
            for idx in range(1, len(networks) + 1):
                params = 'limit=%s' % idx
                request = self.new_list_request(API_RESOURCE, params=params)
                response = self.deserialize(self.fmt,
                                            request.get_response(self.ext_api))
                self.assertEqual(idx, len(response[IP_AVAILS_KEY]))

    def test_usages_query_sorting(self):
        with self.network() as net1, self.network() as net2, \
                self.network() as net3, self.network() as net4:
            networks = (net1, net2, net3, net4)
            network_ids = sorted([net['network']['id'] for net in networks])

            params = 'sort_key=network_id;sort_dir=asc'
            request = self.new_list_request(API_RESOURCE, params=params)
            response = self.deserialize(self.fmt,
                                        request.get_response(self.ext_api))
            res = [net['network_id'] for net in response[IP_AVAILS_KEY]]
            self.assertEqual(network_ids, res)

            params = 'sort_key=network_id;sort_dir=desc'
            request = self.new_list_request(API_RESOURCE, params=params)
            response = self.deserialize(self.fmt,
                                        request.get_response(self.ext_api))
            res = [net['network_id'] for net in response[IP_AVAILS_KEY]]
            self.assertEqual(list(reversed(network_ids)), res)



# Copyright (c) 2015 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.callbacks import events
from neutron_lib.callbacks import exceptions as c_exc
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib.db import api as db_api
from neutron_lib.db import utils as db_utils
from neutron_lib import exceptions as n_exc
from neutron_lib.objects import exceptions as o_exc

from neutron.extensions import rbac as ext_rbac
from neutron.objects import base as base_obj
from neutron.objects import rbac as rbac_obj


class RbacPluginMixin(object):
    """Plugin mixin that implements the RBAC DB operations."""

    object_type_cache = {}
    supported_extension_aliases = ['rbac-policies']

    @db_api.retry_if_session_inactive()
    def create_rbac_policy(self, context, rbac_policy):
        e = rbac_policy['rbac_policy']
        # NOTE(ralonsoh): remove this conversion when "bp/keystone-v3" is
        # widely implemented in all OpenStack projects.
        try:
            e['target_project'] = e.pop('target_tenant')
        except KeyError:
            pass
        try:
            registry.publish(resources.RBAC_POLICY, events.BEFORE_CREATE, self,
                             payload=events.DBEventPayload(
                                 context, request_body=e,
                                 metadata={'object_type': e['object_type']}))
        except c_exc.CallbackFailure as e:
            raise n_exc.InvalidInput(error_message=e)
        rbac_class = (
            rbac_obj.RBACBaseObject.get_type_class_map()[e['object_type']])
        try:
            rbac_args = {'project_id': e['project_id'],
                         'object_id': e['object_id'],
                         'action': e['action'],
                         'target_project': e['target_project']}
            _rbac_obj = rbac_class(context, **rbac_args)
            _rbac_obj.create()
        except o_exc.NeutronDbObjectDuplicateEntry:
            raise ext_rbac.DuplicateRbacPolicy()
        return self._make_rbac_policy_dict(_rbac_obj)

    @staticmethod
    def _make_rbac_policy_dict(entry, fields=None):
        res = {f: entry[f] for f in ('id', 'project_id', 'target_project',
                                     'action', 'object_id')}
        # TODO(ralonsoh): remove once all calls refer to "target_project"
        res['target_tenant'] = res['target_project']
        res['object_type'] = entry.db_model.object_type
        return db_utils.resource_fields(res, fields)

    @db_api.retry_if_session_inactive()
    def update_rbac_policy(self, context, id, rbac_policy):
        pol = rbac_policy['rbac_policy']
        # NOTE(ralonsoh): remove this conversion when "bp/keystone-v3" is
        # widely implemented in all OpenStack projects.
        try:
            pol['target_project'] = pol.pop('target_tenant')
        except KeyError:
            pass
        entry = self._get_rbac_policy(context, id)
        object_type = entry.db_model.object_type
        try:
            registry.publish(resources.RBAC_POLICY, events.BEFORE_UPDATE, self,
                             payload=events.DBEventPayload(
                                 context, request_body=pol,
                                 states=(entry,), resource_id=id,
                                 metadata={'object_type': object_type}))
        except c_exc.CallbackFailure as ex:
            raise ext_rbac.RbacPolicyInUse(object_id=entry.object_id,
                                           details=ex)
        entry.update_fields(pol)
        entry.update()
        return self._make_rbac_policy_dict(entry)

    @db_api.retry_if_session_inactive()
    def delete_rbac_policy(self, context, id):
        entry = self._get_rbac_policy(context, id)
        object_type = entry.db_model.object_type
        try:
            registry.publish(resources.RBAC_POLICY, events.BEFORE_DELETE, self,
                             payload=events.DBEventPayload(
                                 context, states=(entry,), resource_id=id,
                                 metadata={'object_type': object_type}))
        except c_exc.CallbackFailure as ex:
            raise ext_rbac.RbacPolicyInUse(object_id=entry.object_id,
                                           details=ex)
        # make a dict copy because deleting the entry will nullify its
        # object_id link to network
        entry_dict = entry.to_dict()
        entry.delete()
        registry.publish(resources.RBAC_POLICY, events.AFTER_DELETE, self,
                         payload=events.DBEventPayload(
                             context, states=(entry_dict,), resource_id=id,
                             metadata={'object_type': object_type}))
        self.object_type_cache.pop(id, None)

    def _get_rbac_policy(self, context, id):
        object_type = self._get_object_type(context, id)
        rbac_class = rbac_obj.RBACBaseObject.get_type_class_map()[object_type]
        _rbac_obj = rbac_class.get_object(context, id=id)
        if not _rbac_obj:
            raise ext_rbac.RbacPolicyNotFound(id=id, object_type=object_type)
        return _rbac_obj

    @db_api.retry_if_session_inactive()
    def get_rbac_policy(self, context, id, fields=None):
        return self._make_rbac_policy_dict(
            self._get_rbac_policy(context, id), fields=fields)

    @db_api.retry_if_session_inactive()
    def get_rbac_policies(self, context, filters=None, fields=None,
                          sorts=None, limit=None, page_reverse=False):
        pager = base_obj.Pager(sorts, limit, page_reverse)
        filters = filters or {}
        object_types = filters.pop('object_type', None)
        # NOTE(ralonsoh): remove this conversion when "bp/keystone-v3" is
        # widely implemented in all OpenStack projects.
        try:
            filters['target_project'] = filters.pop('target_tenant')
        except KeyError:
            pass
        rbac_classes_to_query = [
            o for t, o in rbac_obj.RBACBaseObject.get_type_class_map().items()
            if not object_types or t in object_types]
        rbac_objs = []
        for rbac_class in rbac_classes_to_query:
            rbac_objs += rbac_class.get_objects(context, _pager=pager,
                                                **filters)
        return [self._make_rbac_policy_dict(_rbac_obj, fields)
                for _rbac_obj in rbac_objs]

    def _get_object_type(self, context, entry_id):
        """Scans all RBAC tables for an ID to figure out the type.

        This will be an expensive operation as the number of RBAC tables grows.
        The result is cached since object types cannot be updated for a policy.
        """
        if entry_id in self.object_type_cache:
            return self.object_type_cache[entry_id]
        for otype, rbac_class in \
                rbac_obj.RBACBaseObject.get_type_class_map().items():
            if rbac_class.count(context, id=entry_id):
                self.object_type_cache[entry_id] = otype
                return otype
        raise ext_rbac.RbacPolicyNotFound(id=entry_id, object_type='unknown')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from neutron.agent import agent_extensions_manager as agent_ext_manager
from neutron.conf.agent import agent_extensions_manager as agent_ext_mgr_config

LOG = log.getLogger(__name__)


L2_AGENT_EXT_MANAGER_NAMESPACE = 'neutron.agent.l2.extensions'


def register_opts(conf):
    agent_ext_mgr_config.register_agent_ext_manager_opts(conf)


class L2AgentExtensionsManager(agent_ext_manager.AgentExtensionsManager):
    """Manage l2 agent extensions. The handle_port and delete_port methods are
       guaranteed to be attributes of each extension because they have been
       marked as abc.abstractmethod in the extensions' abstract class.
    """

    def __init__(self, conf):
        super(L2AgentExtensionsManager,
              self).__init__(conf, L2_AGENT_EXT_MANAGER_NAMESPACE)

    def handle_port(self, context, data):
        """Notify all agent extensions to handle port."""
        for extension in self:
            if hasattr(extension.obj, 'handle_port'):
                extension.obj.handle_port(context, data)
            else:
                LOG.error(
                    "Agent Extension '%(name)s' does not "
                    "implement method handle_port",
                    {'name': extension.name}
                )

    def delete_port(self, context, data):
        """Notify all agent extensions to delete port."""
        for extension in self:
            if hasattr(extension.obj, 'delete_port'):
                extension.obj.delete_port(context, data)
            else:
                LOG.error(
                    "Agent Extension '%(name)s' does not "
                    "implement method delete_port",
                    {'name': extension.name}
                )
